<?php

namespace Drupal\datafield\Plugin\Field\FieldFormatter;

use Drupal\Component\Render\FormattableMarkup;
use Drupal\Core\Datetime\DateFormatterInterface;
use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\Core\Datetime\Entity\DateFormat;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\FormatterBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;
use Drupal\datafield\Plugin\Field\FieldType\DataField as DataFieldItem;

/**
 * Base class for data field formatters.
 */
abstract class Base extends FormatterBase {

  /**
   * Subfield types that can be rendered as a link.
   *
   * @var array
   */
  protected static $linkTypes = ['email', 'telephone', 'uri'];

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings(): array {
    return parent::defaultSettings();
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state): array {
    $formatter_settings_default = [
      'hidden' => FALSE,
      'show_label' => FALSE,
      'link' => FALSE,
      'format_type' => 'medium',
      'thousand_separator' => '',
      'decimal_separator' => '.',
      'scale' => 2,
      'key' => FALSE,
    ];
    $settings = $this->getSetting('formatter_settings');
    $field_settings = $this->getFieldSettings();
    $types = DataFieldItem::subfieldTypes();
    $element = [];
    // General settings.
    foreach ($field_settings["columns"] as $subfield => $item) {
      if (empty($settings[$subfield])) {
        $settings[$subfield] = $formatter_settings_default;
      }
      else {
        $settings[$subfield] += $formatter_settings_default;
      }
      $type = $item['type'];
      $title = $item['name'] . ' - ' . $types[$type];
      if (!empty($field_settings['field_settings']) && $field_settings['field_settings'][$subfield]['list']) {
        $title .= ' (' . $this->t('list') . ')';
      }

      $element['formatter_settings'][$subfield] = [
        '#title' => $title,
        '#type' => 'details',
      ];

      $element['formatter_settings'][$subfield]['link'] = [
        '#type' => 'checkbox',
        '#title' => $this->t('Display as link'),
        '#default_value' => $settings[$subfield]['link'],
        '#weight' => -10,
        '#access' => in_array($type, static::$linkTypes),
      ];

      if ($type == 'datetime_iso8601') {
        $format_types = DateFormat::loadMultiple();
        $date_formatter = static::getDateFormatter();
        $time = new DrupalDateTime();
        $options = [];
        foreach ($format_types as $type => $type_info) {
          $format = $date_formatter->format($time->getTimestamp(), $type);
          $options[$type] = $type_info->label() . ' (' . $format . ')';
        }
        $element['formatter_settings'][$subfield]['format_type'] = [
          '#type' => 'select',
          '#title' => $this->t('Date format'),
          '#description' => $this->t('Choose a format for displaying the date.'),
          '#options' => $options,
          '#default_value' => $settings[$subfield]['format_type'],
        ];
      }
      else {
        $element['formatter_settings'][$subfield]['format_type'] = [
          '#type' => 'value',
          '#default_value' => $settings[$subfield]['format_type'],
        ];
      }

      $element['formatter_settings'][$subfield]['hidden'] = [
        '#type' => 'checkbox',
        '#title' => $this->t('Hidden'),
        '#default_value' => $settings[$subfield]['hidden'],
      ];

      $element['formatter_settings'][$subfield]['show_label'] = [
        '#type' => 'checkbox',
        '#title' => $this->t('Show label'),
        '#default_value' => $settings[$subfield]['show_label'],
      ];

      if (!empty($field_settings['field_settings']) && !empty($field_settings['field_settings'][$subfield]['list'])) {
        $element['formatter_settings'][$subfield]['key'] = [
          '#type' => 'checkbox',
          '#title' => $this->t('Display key instead of label'),
          // @todo Remove the fallback on 5.x.
          '#default_value' => $settings[$subfield]['key'] ?? FALSE,
        ];
      }

      if ($type == 'numeric' || $type == 'float' || $type == 'integer') {
        $options = [
          '' => $this->t('- None -'),
          '.' => $this->t('Decimal point'),
          ',' => $this->t('Comma'),
          ' ' => $this->t('Space'),
          chr(8201) => $this->t('Thin space'),
          "'" => $this->t('Apostrophe'),
        ];
        $element['formatter_settings'][$subfield]['thousand_separator'] = [
          '#type' => 'select',
          '#title' => $this->t('Thousand marker'),
          '#options' => $options,
          '#default_value' => $settings[$subfield]['thousand_separator'],
        ];
      }
      else {
        $element['formatter_settings'][$subfield]['thousand_separator'] = [
          '#type' => 'value',
          '#default_value' => $settings[$subfield]['thousand_separator'],
        ];
      }

      if ($type == 'numeric' || $type == 'float') {
        $element['formatter_settings'][$subfield]['decimal_separator'] = [
          '#type' => 'select',
          '#title' => $this->t('Decimal marker'),
          '#options' => [
            '.' => $this->t('Decimal point'),
            ',' => $this->t('Comma'),
          ],
          '#default_value' => $settings[$subfield]['decimal_separator'],
        ];
        $element['formatter_settings'][$subfield]['scale'] = [
          '#type' => 'number',
          '#title' => $this->t('Scale', [], ['context' => 'decimal places']),
          '#min' => 0,
          '#max' => 10,
          '#default_value' => $settings[$subfield]['scale'],
          '#description' => $this->t('The number of digits to the right of the decimal.'),
        ];
      }
      else {
        $element['formatter_settings'][$subfield]['decimal_separator'] = [
          '#type' => 'value',
          '#default_value' => $settings[$subfield]['decimal_separator'],
        ];
        $element['formatter_settings'][$subfield]['scale'] = [
          '#type' => 'value',
          '#default_value' => $settings[$subfield]['scale'],
        ];
      }

    }

    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary(): array {
    $settings = $this->getSetting('formatter_settings');
    $field_settings = $this->getFieldSettings();

    $subfield_types = DataFieldItem::subfieldTypes();

    $summary = [];
    foreach ($field_settings["columns"] as $subfield => $item) {
      $subfield_type = $item['type'];
      if (empty($settings[$subfield]) || empty($field_settings["field_settings"])) {
        continue;
      }
      $summary[] = new FormattableMarkup(
        '<b>@subfield - @subfield_type@list</b>',
        [
          '@subfield' => $field_settings["field_settings"][$subfield]["label"],
          '@subfield_type' => strtolower($subfield_types[$subfield_type]),
          '@list' => $field_settings["field_settings"][$subfield]["list"] ? ' (' . $this->t('list') . ')' : '',
        ]
      );
      if (isset($settings[$subfield]['format_type']) && $subfield_type == 'datetime_iso8601') {
        $summary[] = $this->t('Date format: @format', ['@format' => $settings[$subfield]['format_type']]);
      }
      if (isset($settings[$subfield]['link']) && in_array($subfield_type, static::$linkTypes)) {
        $summary[] = $this->t('Link: @value', ['@value' => $settings[$subfield]['link'] ? $this->t('yes') : $this->t('no')]);
      }
      if (isset($settings[$subfield]['hidden'])) {
        $summary[] = $this->t('Hidden: @value', ['@value' => $settings[$subfield]['hidden'] ? $this->t('yes') : $this->t('no')]);
      }
      if (!empty($field_settings[$subfield]["list"])) {
        // @todo Remove the fallback in 5.x.
        $display_key = $settings[$subfield]['key'] ?? FALSE;
        $summary[] = $this->t('Display key: @value', ['@value' => $display_key ? $this->t('yes') : $this->t('no')]);

      }
      if ($subfield_type == 'numeric' || $subfield_type == 'float' || $subfield_type == 'integer') {
        $summary[] = $this->t('Number format: @format', ['@format' => $this->numberFormat($subfield, 1234.1234567890)]);
      }
    }

    return $summary;
  }

  /**
   * {@inheritdoc}
   */
  public function view(FieldItemListInterface $items, $langcode = NULL): array {
    $elements = [];
    if (count($items) > 0) {
      // A field may appear multiple times in a single view. Since items are
      // passed by reference we need to ensure they are processed only once.
      $items = clone $items;
      $this->prepareItems($items);
      $elements = parent::view($items, $langcode);
    }
    return $elements;
  }

  /**
   * Prepare field items.
   *
   * @param \Drupal\Core\Field\FieldItemListInterface $items
   *   List of field items.
   */
  protected function prepareItems(FieldItemListInterface $items): void {

    $field_settings = $this->getFieldSettings();
    $settings = $this->getSettings();

    foreach ($items as $delta => $item) {
      foreach ($field_settings["columns"] as $subfield => $storage) {

        if ($settings['formatter_settings'][$subfield]['hidden']) {
          $item->{$subfield} = NULL;
        }
        else {

          $type = $storage['type'];

          if ($type == 'boolean') {
            $item->{$subfield} = $field_settings[$subfield][$item->{$subfield} ? 'on_label' : 'off_label'];
          }

          // Empty string should already be converted into NULL.
          // @see Drupal\datafield\Plugin\Field\FieldWidget\DataField::massageFormValues()
          if ($item->{$subfield} === NULL) {
            continue;
          }

          if ($type == 'numeric' || $type == 'float' || $type == 'integer') {
            $item->{$subfield} = $this->numberFormat($subfield, $item->{$subfield});
          }

          if (in_array($type, ['datetime_iso8601', 'date']) &&
            $item->{$subfield} &&
            in_array($storage["datetime_type"],['datetime', 'timestamp', 'date', 'week', 'month', 'year'])) {
            // We follow the same principles as Drupal Core.
            // In the case of a datetime subfield, the date must be parsed using
            // the storage time zone and converted to the user's time zone while
            // a date-only field should have no timezone conversion performed.
            $timezone = $storage['datetime_type'] === 'datetime' ?
              date_default_timezone_get() : DataFieldItem::DATETIME_STORAGE_TIMEZONE;
            if ($type == 'datetime_iso8601') {
              if (in_array($storage["datetime_type"], ['datetime', 'date'])) {
                $timestamp = $items[$delta]->createDate($subfield)
                  ->getTimestamp();
              }
              elseif ($storage["datetime_type"] == 'timestamp' && is_numeric($item->{$subfield})) {
                $timestamp = $item->{$subfield};
              }
              else {
                $timestamp = strtotime($item->{$subfield});
              }
            }

            if ($type == 'date') {
              if ($storage["datetime_type"] == 'timestamp' && is_numeric($item->{$subfield})) {
                $timestamp = $item->{$subfield};
              }
              else {
                $timestamp = strtotime($item->{$subfield});
              }
            }
            $date_formatter = static::getDateFormatter();
            $formatType = $settings['formatter_settings'][$subfield]['format_type'];
            $date = $date_formatter->format($timestamp, $formatType, '', $timezone);
            if ($storage["datetime_type"] == 'timestamp' && is_numeric($item->{$subfield})) {
              $date = \Drupal::service('date.formatter')->format(
                $timestamp, $formatType
              );
            }
            $item->{$subfield} = [
              '#theme' => 'time',
              '#text' => $date,
              '#html' => FALSE,
              '#attributes' => [
                'datetime' => $date_formatter->format($timestamp, 'custom', 'Y-m-d\TH:i:s') . 'Z',
              ],
              '#cache' => [
                'contexts' => [
                  'timezone',
                ],
              ],
            ];
          }

          $original_value[$subfield] = $item->{$subfield};
          if (!empty($field_settings["field_settings"]) && $field_settings["field_settings"][$subfield]['list']) {
            // @todo Remove the fallback in 5.x.
            $display_key = $settings['formatter_settings'][$subfield]['key'] ?? FALSE;
            if (!$display_key) {
              // Replace the value with its label if possible.
              $item->{$subfield} = $field_settings["field_settings"][$subfield]['allowed_values'][$item->{$subfield}] ?? NULL;
            }
          }

          if (!empty($settings['formatter_settings']) && !empty($settings['formatter_settings'][$subfield]['link'])) {
            $value = isset($original_value) ? $original_value[$subfield] : $item->{$subfield};
            switch ($type) {
              case 'email':
                $item->{$subfield} = [
                  '#type' => 'link',
                  '#title' => $item->{$subfield},
                  '#url' => Url::fromUri('mailto:' . $value),
                ];
                break;

              case 'telephone':
                $item->{$subfield} = [
                  '#type' => 'link',
                  '#title' => $item->{$subfield},
                  '#url' => Url::fromUri('tel:' . rawurlencode(preg_replace('/\s+/', '', $value))),
                  '#options' => ['external' => TRUE],
                ];
                break;

              case 'uri':
                $item->{$subfield} = [
                  '#type' => 'link',
                  '#title' => $item->{$subfield},
                  '#url' => Url::fromUri($value),
                  '#options' => ['external' => TRUE],
                ];
                break;

            }
          }

        }

      }
      $items[$delta] = $item;
    }
  }

  /**
   * Formats a number.
   */
  protected function numberFormat(string $subfield, string $number): string {
    $settings = $this->getSetting('formatter_settings')[$subfield];
    if ($this->getFieldSetting('columns')[$subfield]['type'] == 'integer') {
      $settings['scale'] = 0;
    }
    return number_format($number, $settings['scale'], $settings['decimal_separator'], $settings['thousand_separator']);
  }

  /**
   * Returns date formatter.
   */
  protected static function getDateFormatter(): DateFormatterInterface {
    return \Drupal::service('date.formatter');
  }

}
